/* -*- c -*- */

#define _GNU_SOURCE
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <stdbool.h>
#include <string.h>
#include <math.h>

#include <coff.h>

#include <bw/bw.h>
#include <bw/boio.h>
#include <bw/max.h>
#include <bw/min.h>

#include "utils.h"

#define SCLASS_NAMES(class) [C_##class] = __bw_stringify(class)
static const char *const sclass_names[256] = {
    SCLASS_NAMES(NULL),
    SCLASS_NAMES(AUTO),
    SCLASS_NAMES(EXT),
    SCLASS_NAMES(STAT),
    SCLASS_NAMES(REG),
    SCLASS_NAMES(EXTDEF),
    SCLASS_NAMES(LABEL),
    SCLASS_NAMES(ULABEL),
    SCLASS_NAMES(MOS),
    SCLASS_NAMES(ARG),
    SCLASS_NAMES(STRTAG),
    SCLASS_NAMES(MOU),
    SCLASS_NAMES(UNTAG),
    SCLASS_NAMES(TPDEF),
    SCLASS_NAMES(USTATIC),
    SCLASS_NAMES(ENTAG),
    SCLASS_NAMES(MOE),
    SCLASS_NAMES(REGPARM),
    SCLASS_NAMES(FIELD),
    SCLASS_NAMES(AUTOARG),
    SCLASS_NAMES(LASTENT),
    SCLASS_NAMES(BLOCK),
    SCLASS_NAMES(FCN),
    SCLASS_NAMES(EOS),
    SCLASS_NAMES(FILE),
    SCLASS_NAMES(LINE),
    SCLASS_NAMES(ALIAS),
    SCLASS_NAMES(HIDDEN),
    SCLASS_NAMES(EFCN)
};
#undef SCLASS_NAMES

static const char *const f_flags_names[] = {
    [ 0] = "RELFLG",
    [ 1] = "EXEC",
    [ 2] = "LNNO",
    [ 3] = "LSYMS",
    [ 7] = "AR16WR",
    [ 8] = "AR32WR",
    [ 9] = "AR32W",
    [12] = "DYNLOAD",
    [13] = "SHROBJ"
};

#define S_FLAG_NAME(flag) case STYP_##flag: return __bw_stringify(flag)
static inline const char *
s_flags_names(uint16_t s_flags)
{
    switch (s_flags) {
        S_FLAG_NAME(REG);
        S_FLAG_NAME(DSECT);
        S_FLAG_NAME(NOLOAD);
        S_FLAG_NAME(GROUP);
        S_FLAG_NAME(PAD);
        S_FLAG_NAME(COPY);
        S_FLAG_NAME(TEXT);
        S_FLAG_NAME(DATA);
        S_FLAG_NAME(BSS);
        case S_NEWFCN: return "NEWFCN";
        S_FLAG_NAME(INFO);
        S_FLAG_NAME(OVER);
        S_FLAG_NAME(LIB);
        S_FLAG_NAME(MERGE);
        S_FLAG_NAME(REVERSE_PAD);
        S_FLAG_NAME(LIT);
        default: return "UNKNOWN";
    }
}
#undef S_FLAG_NAME

#define R_TYPE_NAME(type) case R_##type: return __bw_stringify(type)
static inline const char *
r_type_names(uint16_t r_type)
{
    switch (r_type) {
        R_TYPE_NAME(13REL);
        R_TYPE_NAME(13ABS);
        R_TYPE_NAME(19REL);
        R_TYPE_NAME(19ABS);
        R_TYPE_NAME(32DAT);
        R_TYPE_NAME(GETHI);
        default: return "UNKNOWN";
    }
}
#undef R_TYPE_NAME


struct coff_file *
read_coff(const char *name)
{
    struct coff_file *coff = NULL;
    FILE *f;
    uint16_t ui16;
    int32_t i32;
    size_t ret;
    char errstr[512];
    int succ;

    f = fopen(name, "r");
    if (!f) {
        fprintf(stderr, "%s:%s: error: can't open file: %s\n",
            __func__, name, strerror_r(errno, errstr, sizeof(errstr)));
        goto out;
    }

    coff = calloc(1, sizeof(*coff));
    if (!coff) {
        fprintf(stderr, "%s:%s: error: not enough memory\n", __func__, name);
        goto out_fclose;
    }

    coff->name = name;

    ret = coff_filehdr_fread(&coff->filehdr, 1, f);
    if (ret != 1) {
        fprintf(stderr, "%s:%s: error: can't read filehdr: %s\n",
            __func__, name,
            feof(f) ? "Unexpected end of file"
                    : strerror_r(ferror(f), errstr, sizeof(errstr)));
        goto out_free_coff;
    }

    if (coff->filehdr.f_opthdr != AOUTHDR_SIZE) {
        fprintf(stderr,
            "%s:%s: skip/ignore unknown optional header of size %"PRIu16"\n",
            __func__, name, coff->filehdr.f_opthdr
        );
        succ = fseek(f, coff->filehdr.f_opthdr, SEEK_CUR);
        if (succ < 0) {
            fprintf(stderr,
                "%s:%s: error: can't seek behind f_opthdr: %s\n",
                __func__, name,
                feof(f) ? "Unexpected end of file"
                        : strerror_r(ferror(f), errstr, sizeof(errstr))
            );
            goto out_free_coff;
        }
    } else {
        coff->aouthdr = calloc(1, sizeof(*coff->aouthdr));
        if (!coff->aouthdr) {
            fprintf(stderr, "%s:%s: error: not enough memory\n",
                __func__, name);
            goto out_free_coff;
        }

        ret = coff_aouthdr_fread(coff->aouthdr, 1, f);
        if (ret != 1) {
            fprintf(stderr,
                "%s:%s: error: can't read aouthdr: %s\n",
                __func__, name,
                feof(f) ? "Unexpected end of file"
                        : strerror_r(ferror(f), errstr, sizeof(errstr))
            );
            goto out_free_aouthdr;
        }
    }

    /* section headers */
    if (coff->filehdr.f_nscns) {
        coff->sections = calloc(coff->filehdr.f_nscns, sizeof(*coff->sections));
        if (!coff->sections) {
            fprintf(stderr, "%s:%s: error: not enough memory\n",
                __func__, name);
            goto out_free_aouthdr;
        }
    }

    for (ui16 = 0; ui16 < coff->filehdr.f_nscns; ui16++) {
        struct coff_section *section = &coff->sections[ui16];

        ret = coff_scnhdr_fread(&section->head, 1, f);
        if (ret != 1) {
            fprintf(stderr,
                "%s:%s: error: can't read section[%"PRIu16"] header: %s\n",
                __func__, name, ui16 + 1,
                feof(f) ? "Unexpected end of file"
                        : strerror_r(ferror(f), errstr, sizeof(errstr))
            );
            goto out_free_sections;
        }
    }

    /* section data */
    for (ui16 = 0; ui16 < coff->filehdr.f_nscns; ui16++) {
        struct coff_section *section = &coff->sections[ui16];

        /* reduce size from byte count to word count */
        section->head.s_size /= sizeof(int32_t);

        if (section->head.s_flags != STYP_BSS) {
            section->data
                = calloc(section->head.s_size, sizeof(*section->data));
            if (!section->data) {
                fprintf(stderr, "%s:%s: error: not enough memory\n",
                    __func__, name);
                goto out_free_sections;
            }

            if (ftell(f) != (long)section->head.s_scnptr)
                fprintf(stderr,
                    "%s:%s: warning: wrong scnptr for section[%"PRIu16"]\n",
                    __func__, name, ui16 + 1
                );

            ret = bw_fread_be32(section->data, section->head.s_size, f);

            if (ret != (size_t)section->head.s_size) {
                fprintf(stderr,
                    "%s:%s: error: can't read section[%"PRIu16"] data: %s\n",
                    __func__, name, ui16 + 1,
                    feof(f) ? "Unexpected end of file"
                            : strerror_r(ferror(f), errstr, sizeof(errstr))
                );
                goto out_free_sections;
            }

        }
    }

    /* relocations */
    for (ui16 = 0; ui16 < coff->filehdr.f_nscns; ui16++) {
        struct coff_section *section = &coff->sections[ui16];
        uint16_t uj16;

        if (ftell(f) != (long)section->head.s_relptr)
            fprintf(stderr,
                "%s:%s: warning: wrong relptr for section[%"PRIu16"]\n",
                __func__, name, ui16 + 1
            );

        if (section->head.s_nreloc) {
            section->relocs
                = calloc(section->head.s_nreloc, sizeof(*section->relocs));
            if (!section->relocs) {
                fprintf(stderr, "%s:%s: error: not enough memory\n",
                    __func__, name);
                goto out_free_sections;
            }

            for (uj16 = 0; uj16 < section->head.s_nreloc; uj16++) {
                struct reloc *reloc = &section->relocs[uj16];

                ret = coff_reloc_fread(reloc, 1, f);
                if (ret != 1) {
                    fprintf(stderr,
                        "%s:%s: error: can't read "
                        "section[%"PRIu16"].reloc[%"PRIu16"]: %s\n",
                        __func__, name, ui16 + 1, uj16,
                        feof(f) ? "Unexpected end of file"
                                : strerror_r(ferror(f), errstr, sizeof(errstr))
                    );
                    goto out_free_sections;
                }

            }

        }
    }

    /* line numbers */
    for (ui16 = 0; ui16 < coff->filehdr.f_nscns; ui16++) {
        struct coff_section *section = &coff->sections[ui16];
        uint16_t uj16;

        if (ftell(f) != (long)section->head.s_lnnoptr)
            fprintf(stderr,
                "%s:%s: warning: wrong lnnoptr for section %"PRIu16"\n",
                __func__, name, ui16 + 1
            );

        if (section->head.s_nlnno) {
            section->linenos
                = calloc(section->head.s_nlnno, sizeof(*section->linenos));
            if (!section->linenos) {
                fprintf(stderr, "%s:%s: error: not enough memory\n",
                    __func__, name);
                goto out_free_sections;
            }

            for (uj16 = 0; uj16 < section->head.s_nlnno; uj16++) {
                struct lineno *lineno = &section->linenos[uj16];

                ret = coff_lineno_fread(lineno, 1, f);
                if (ret != 1) {
                    fprintf(stderr,
                        "%s:%s: error: can't read "
                        "section[%"PRIu16"].lnno[%"PRIu16"]: %s\n",
                        __func__, name, ui16 + 1, uj16,
                        feof(f) ? "Unexpected end of file"
                                : strerror_r(ferror(f), errstr, sizeof(errstr))
                    );
                    goto out_free_sections;
                }

            }

        }
    }

    if (coff->filehdr.f_nsyms) {
        coff->syments = calloc(coff->filehdr.f_nsyms, sizeof(*coff->syments));
        if (!coff->syments) {
            fprintf(stderr, "%s:%s: error: not enough memory\n",
                __func__, name);
            goto out_free_sections;
        }
    }

    if (ftell(f) != (long)coff->filehdr.f_symptr)
        fprintf(stderr, "%s:%s: warning: wrong symptr\n", __func__, name);

    for (i32 = 0; i32 < coff->filehdr.f_nsyms; i32++) {
        struct coff_syment *syment = &coff->syments[i32];

        ret = coff_syment_fread(&syment->syment, 1, f);
        if (ret != 1) {
            fprintf(stderr,
                "%s:%s: error: can't read syment[%"PRIu16"]: %s\n",
                __func__, name, i32,
                feof(f) ? "Unexpected end of file"
                        : strerror_r(ferror(f), errstr, sizeof(errstr))
            );
            goto out_free_syments;
        }

        /* skip auxiliary entries */
        if (syment->syment.n_numaux) {
            succ = fseek(f, syment->syment.n_numaux * AUXENT_SIZE, SEEK_CUR);
            if (succ < 0) {
                fprintf(stderr,
                    "%s:%s: error: can't seek behind auxiliary entries for syment[%"PRIu16"]: %s\n",
                    __func__, name, i32,
                    feof(f) ? "Unexpected end of file"
                            : strerror_r(ferror(f), errstr, sizeof(errstr))
                );
                goto out_free_syments;
            }
        }

    }

    ret = bw_fread_be32(&coff->str_tab_len, 1, f);
    if (ret != 1) {
        fprintf(stderr, "%s:%s: error: can't read string table length: %s\n",
            __func__, name,
            feof(f) ? "Unexpected end of file"
                    : strerror_r(ferror(f), errstr, sizeof(errstr)));
        goto out_free_syments;
    }

    coff->str_tab = calloc(coff->str_tab_len, sizeof(*coff->str_tab));
    if (!coff->str_tab) {
        fprintf(stderr, "%s:%s: error: not enough memory\n", __func__, name);
        goto out_free_syments;
    }

    ret = fread(coff->str_tab + 4, sizeof(char), coff->str_tab_len - 4, f);
    if (ret != (size_t)(coff->str_tab_len - 4)) {
        fprintf(stderr, "%s:%s: error: can't read string table: %s\n",
            __func__, name,
            feof(f) ? "Unexpected end of file"
                    : strerror_r(ferror(f), errstr, sizeof(errstr)));
        goto out_free_str_tab;
    }

    if (feof(f)) {
        fprintf(stderr, "%s:%s: error: file too great\n", __func__, name);
        goto out_free_str_tab;
    }

    /* set syment->name to real name */
    for (i32 = 0; i32 < coff->filehdr.f_nsyms; i32++) {
        struct coff_syment *syment = &coff->syments[i32];

        if (syment->syment.n_zeroes) {
            syment->name = syment->syment.n_name;
        } else {
            syment->name = &coff->str_tab[syment->syment.n_offset];
        }

    }

    goto out_fclose;

out_free_str_tab:
    if (coff->str_tab)
        free(coff->str_tab);
    coff->str_tab = NULL;
out_free_syments:
    if (coff->syments)
        free(coff->syments);
    coff->syments = NULL;
out_free_sections:
    if (coff->sections) {
        for (ui16 = 0; ui16 < coff->filehdr.f_nscns; ui16++) {
            struct coff_section *section = &coff->sections[ui16];

            if (section->data)
                free(section->data);
            section->data = NULL;
            if (section->relocs)
                free(section->relocs);
            section->relocs = NULL;
            if (section->linenos)
                free(section->linenos);
            section->linenos = NULL;

        }
        free(coff->sections);
        coff->sections = NULL;
    }
out_free_aouthdr:
    if (coff->aouthdr)
        free(coff->aouthdr);
    coff->aouthdr = NULL;
out_free_coff:
    free(coff);
    coff = NULL;
out_fclose:
    fclose(f);
out:
    return coff;
}

void
free_coff(struct coff_file *coff)
{
    uint16_t ui16;

    if (!coff)
        return;

    if (coff->str_tab)
        free(coff->str_tab);
    coff->str_tab = NULL;

    if (coff->syments)
        free(coff->syments);
    coff->syments = NULL;

    if (coff->sections) {
        for (ui16 = 0; ui16 < coff->filehdr.f_nscns; ui16++) {
            struct coff_section *section = &coff->sections[ui16];

            if (section->data)
                free(section->data);
            section->data = NULL;
            if (section->relocs)
                free(section->relocs);
            section->relocs = NULL;
            if (section->linenos)
                free(section->linenos);
            section->linenos = NULL;

        }
        free(coff->sections);
        coff->sections = NULL;
    }

    if (coff->aouthdr)
        free(coff->aouthdr);
    coff->aouthdr = NULL;

    free(coff);
}

void print_coff(const struct coff_file *coff, int flags)
{
    uint16_t ui16;
    int32_t i32;

    if (!coff)
        return;

    if (flags & COFF_FILEHDR) {
        int bit;
        printf(
            "coff header:\n"
            "  f_magic:         %04"PRIx16"\n"
            "  f_nscns:         %04"PRId16"\n"
            "  f_timdat:        %"PRId32"\n"
            "  f_symptr:        %"PRId32"\n"
            "  f_nsyms:         %"PRId32"\n"
            "  f_opthdr:        %"PRId16"\n"
            "  f_flags:         %04"PRIx16"",
            coff->filehdr.f_magic,
            coff->filehdr.f_nscns,
            coff->filehdr.f_timdat,
            coff->filehdr.f_symptr,
            coff->filehdr.f_nsyms,
            coff->filehdr.f_opthdr,
            coff->filehdr.f_flags
        );
        if (coff->filehdr.f_flags) {
            int first = 1;
            printf(" (");
            for (bit = 0; bit < 16; bit++) {
                if ((1 << bit) & coff->filehdr.f_flags) {
                    printf("%s%s",
                        first ? "" : " | ",
                        f_flags_names[bit] ?: "UNKNOWN");
                    first = 0;
                }
            }
            printf(")");
        }
        printf("\n");
    }

    if (flags & COFF_AOUTHDR && coff->filehdr.f_opthdr == AOUTHDR_SIZE) {

        printf(
            "a.out header:\n"
            "  magic:           %04"PRIx16"\n"
            "  vstamp:          %04"PRId16"\n"
            "  tsize:           %"PRId32"\n"
            "  dsize:           %"PRId32"\n"
            "  bsize:           %"PRId32"\n"
            "  entry:           %"PRId32"\n"
            "  text_start:      %"PRId32"\n"
            "  data_start:      %"PRId32"\n",
            coff->aouthdr->magic,
            coff->aouthdr->vstamp,
            coff->aouthdr->tsize,
            coff->aouthdr->dsize,
            coff->aouthdr->bsize,
            coff->aouthdr->entry,
            coff->aouthdr->text_start,
            coff->aouthdr->data_start
        );
    }

    if (flags & COFF_SCNS) {

        for (ui16 = 0; ui16 < coff->filehdr.f_nscns; ui16++) {
            struct coff_section *section = &coff->sections[ui16];
            uint16_t uj16;

            printf(
                "section header: %s\n"
                "  s_paddr:         %08"PRIx32"\n"
                "  s_vaddr:         %08"PRIx32"\n"
                "  s_size:          %"PRId32"\n"
                "  s_scnptr:        %"PRId32"\n"
                "  s_relptr:        %"PRId32"\n"
                "  s_lnnoptr:       %"PRId32"\n"
                "  s_nreloc:        %"PRId16"\n"
                "  s_nlnno:         %"PRId16"\n"
                "  s_flags:         %08"PRIx32" [%s]\n",
                section->head.s_name,
                section->head.s_paddr,
                section->head.s_vaddr,
                section->head.s_size,
                section->head.s_scnptr,
                section->head.s_relptr,
                section->head.s_lnnoptr,
                section->head.s_nreloc,
                section->head.s_nlnno,
                section->head.s_flags,
                s_flags_names(section->head.s_flags)
            );
            

            if (flags & COFF_SCNSDATA
                && section->head.s_size
                && section->head.s_flags != STYP_BSS) {
                FILE *hexdump = popen(
                    "hexdump -v -e '\"    \" 8/4 \"%08x \" \"\n\"'", "w");
                if (hexdump) {
                    printf("  section data:\n");
                    fflush(stdout);
                    fwrite(section->data, sizeof(*section->data),
                        section->head.s_size, hexdump);
                    pclose(hexdump);
                    fflush(stdout);
                    usleep(100);
                }
            }

            if (flags & COFF_RELOC && section->head.s_nreloc) {
                int reloc_maxcol[3] = {
                    strlen("entry"), strlen("symndx"), strlen("offset")
                };
                struct reloc relocmax;
                memset(&relocmax, 0, sizeof(relocmax));

                printf("  relocation directives (%"PRIu16" entries):\n",
                    section->head.s_nreloc);

                for (uj16 = 0; uj16 < section->head.s_nreloc; uj16++) {
                    struct reloc *reloc = &section->relocs[uj16];

                    relocmax.r_symndx = i32max(
                        relocmax.r_symndx,
                        reloc->r_symndx
                    );
                    relocmax.r_offset = u32max(
                        relocmax.r_offset,
                        reloc->r_offset
                    );
                }
                reloc_maxcol[0] = max(reloc_maxcol[0],
                    (int)floor(log10(section->head.s_nreloc)));
                reloc_maxcol[1] = max(reloc_maxcol[1],
                    (int)floor(log10(relocmax.r_symndx)));
                reloc_maxcol[2] = max(reloc_maxcol[2],
                    (int)floor(log10(relocmax.r_offset)));

                printf(
                    "    %*s: vaddr    %*s %*s type\n",
                    reloc_maxcol[0], "entry",
                    reloc_maxcol[1], "symndx",
                    reloc_maxcol[2], "offset"
                );

                for (uj16 = 0; uj16 < section->head.s_nreloc; uj16++) {
                    struct reloc *reloc = &section->relocs[uj16];
                    printf(
                        "    %*"PRIu16": "
                        "%08"PRIx32" "
                        "%*"PRId32" "
                        "%*"PRId32" "
                        "%-s\n",
                        reloc_maxcol[0], uj16,
                        reloc->r_vaddr,
                        reloc_maxcol[1], reloc->r_symndx,
                        reloc_maxcol[2], reloc->r_offset,
                        r_type_names(reloc->r_type)
                    );
                }

            }

            if (flags & COFF_LNNO && section->head.s_nlnno) {
                int lineno_maxcol[2] = {
                    strlen("entry"),
                    strlen("lnno")
                };
                struct lineno linenomax;
                memset(&linenomax, 0, sizeof(linenomax));

                for (uj16 = 0; uj16 < section->head.s_nlnno; uj16++) {
                    struct lineno *lineno = &section->linenos[uj16];
                    linenomax.l_lnno = i32max(linenomax.l_lnno, lineno->l_lnno);
                }
                lineno_maxcol[0] = max(
                    lineno_maxcol[0], (int)floor(log10(section->head.s_nlnno))
                );
                lineno_maxcol[1] = max(
                    lineno_maxcol[1], (int)floor(log10(linenomax.l_lnno))
                );

                printf(
                    "    %*s: paddr    %*s\n",
                    lineno_maxcol[0], "entry",
                    lineno_maxcol[1], "lnno"
                );

                printf("  line numbers: (%"PRIu16" entries)\n",
                    section->head.s_nlnno);

                for (uj16 = 0; uj16 < section->head.s_nlnno; uj16++) {
                    struct lineno *lineno = &section->linenos[uj16];
                    printf(
                        "    %*"PRIu16": "
                        "%08"PRIx32" "
                        "%*"PRId32"\n",
                        lineno_maxcol[0], uj16,
                        lineno->l_paddr,
                        lineno_maxcol[1], lineno->l_lnno
                    );
                }

            }

        }

    }

    if (flags & COFF_SYMTAB) {
        int symtab_maxcol[2] = {
            strlen("entry"),
            strlen("value")
        };

        if (coff->filehdr.f_nsyms) {
            struct syment symentmax;
            memset(&symentmax, 0, sizeof(symentmax));

            printf("symbol table (%"PRId32" entries):\n", coff->filehdr.f_nsyms);

            for (i32 = 0; i32 < coff->filehdr.f_nsyms; i32++) {
                struct coff_syment *syment = &coff->syments[i32];
                symentmax.n_value  = i32max(symentmax.n_value,  syment->syment.n_value);
            }
            symtab_maxcol[0] = max(
                symtab_maxcol[0], (int)floor(log10(coff->filehdr.f_nsyms)));
            symtab_maxcol[1] = max(
                symtab_maxcol[1], (int)floor(log10(symentmax.n_value)));

            printf(
                "  %*s: %*s scnum type sclass  numaux  name\n",
                symtab_maxcol[0], "entry",
                symtab_maxcol[1], "value"
            );
        }

        for (i32 = 0; i32 < coff->filehdr.f_nsyms; i32++) {
            struct coff_syment *syment = &coff->syments[i32];

            printf(
                "  %*"PRId32": "
                "%*"PRId32" "
                "%5"PRId16" "
                "%04"PRIx16" "
                "%-7s "
                "%6"PRIu8" "
                "%c%s\n",
                symtab_maxcol[0], i32,
                symtab_maxcol[1], syment->syment.n_value,
                syment->syment.n_scnum,
                syment->syment.n_type,
                sclass_names[syment->syment.n_sclass] ?: "UNKNOWN",
                syment->syment.n_numaux,
                syment->syment.n_zeroes ? ' ' : '*',
                syment->name
            );
        }

    }

}

int
compare_coff(struct coff_file *coff1, struct coff_file *coff2, bool keep_going)
{
    uint16_t ui16;
    int32_t i32, nsysms_min;
    int c;
    uint16_t nscns_min;

#define coff_cmp(member)                                        \
({                                                              \
    int c = (coff1->member > coff2->member)                     \
          - (coff1->member < coff2->member);                    \
    if (c) {                                                    \
        printf("%s %s differ in %s\n",                          \
            coff1->name, coff2->name, __bw_stringify(member));  \
        if (!keep_going)                                        \
            return c;                                           \
    }                                                           \
})

#define coff_cmp1(member0, idx0, member1)                                   \
({                                                                          \
    int c = (coff1->member0[idx0].member1 > coff2->member0[idx0].member1)   \
          - (coff1->member0[idx0].member1 < coff2->member0[idx0].member1);  \
    if (c) {                                                                \
        printf("%s %s differ in %s[%u].%s\n",                               \
            coff1->name, coff2->name, __bw_stringify(member0),              \
            idx0, __bw_stringify(member1));                                 \
        if (!keep_going)                                                    \
            return c;                                                       \
    }                                                                       \
})

#define coff_cmp2(member0, idx0, member1, idx1, member2)    \
({                                                          \
    int c = (coff1->member0[idx0].member1[idx1].member2     \
             > coff2->member0[idx0].member1[idx1].member2)  \
          - (coff1->member0[idx0].member1[idx1].member2     \
             < coff2->member0[idx0].member1[idx1].member2); \
    if (c) {                                                \
        printf("%s %s differ in %s[%u].%s[%u].%s\n",        \
            coff1->name, coff2->name,                       \
            __bw_stringify(member0), idx0,                  \
            __bw_stringify(member1), idx1,                  \
            __bw_stringify(member2));                       \
        if (!keep_going)                                    \
            return c;                                       \
    }                                                       \
})

    if (!coff1 || !coff2)
        return 0;

    coff_cmp(filehdr.f_magic);
    coff_cmp(filehdr.f_nscns);
    coff_cmp(filehdr.f_timdat);
    coff_cmp(filehdr.f_symptr);
    coff_cmp(filehdr.f_nsyms);
    coff_cmp(filehdr.f_opthdr);
    coff_cmp(filehdr.f_flags);

    if (coff1->aouthdr) {
        coff_cmp(aouthdr->magic);
        coff_cmp(aouthdr->vstamp);
        coff_cmp(aouthdr->tsize);
        coff_cmp(aouthdr->dsize);
        coff_cmp(aouthdr->bsize);
        coff_cmp(aouthdr->entry);
        coff_cmp(aouthdr->text_start);
        coff_cmp(aouthdr->data_start);
    }

    nscns_min = u16min(coff1->filehdr.f_nscns, coff2->filehdr.f_nscns);
    for (ui16 = 0; ui16 < nscns_min; ui16++) {
        struct coff_section *section1 = &coff1->sections[ui16];
        struct coff_section *section2 = &coff2->sections[ui16];
        uint16_t uj16;
        uint16_t nreloc_min, nlnno_min;
        int32_t size_min;

        c = strcmp(section1->head.s_name, section2->head.s_name);
        if (c) {
            printf("%s %s differ in section name %d\n",
                coff1->name, coff2->name, ui16);
            if (!keep_going)
                return c;
        }

        coff_cmp1(sections, ui16, head.s_paddr);
        coff_cmp1(sections, ui16, head.s_vaddr);
        coff_cmp1(sections, ui16, head.s_size);
        coff_cmp1(sections, ui16, head.s_scnptr);
        coff_cmp1(sections, ui16, head.s_relptr);
        coff_cmp1(sections, ui16, head.s_lnnoptr);
        coff_cmp1(sections, ui16, head.s_nreloc);
        coff_cmp1(sections, ui16, head.s_nlnno);
        coff_cmp1(sections, ui16, head.s_flags);

        size_min = i32min(section1->head.s_size, section2->head.s_size);
        if (size_min && section1->head.s_flags != STYP_BSS) {
            for (i32 = 0; i32 < size_min; i32++) {
                c = (section1->data[i32] > section2->data[i32])
                  - (section1->data[i32] < section2->data[i32]);
                if (c) {
                    printf(
                        "%s %s differ in sections[%u].data[%d]: "
                        "%08"PRIx32"|%08"PRIx32"\n",
                        coff1->name, coff2->name, ui16, i32,
                        section1->data[i32], section2->data[i32]
                    );
                    if (!keep_going)
                        return c;
                }
            }
        }

        nreloc_min = u16min(section1->head.s_nreloc, section2->head.s_nreloc);
        if (nreloc_min) {

            for (uj16 = 0; uj16 < nreloc_min; uj16++) {
                coff_cmp2(sections, ui16, relocs, uj16, r_vaddr);
                coff_cmp2(sections, ui16, relocs, uj16, r_symndx);
                coff_cmp2(sections, ui16, relocs, uj16, r_offset);
                coff_cmp2(sections, ui16, relocs, uj16, r_type);
                coff_cmp2(sections, ui16, relocs, uj16, r_stuff);
            }

        }

        nlnno_min = u16min(section1->head.s_nlnno, section2->head.s_nlnno);
        if (nlnno_min) {

            for (uj16 = 0; uj16 < nlnno_min; uj16++) {
                coff_cmp2(sections, ui16, linenos, uj16, l_paddr);
                coff_cmp2(sections, ui16, linenos, uj16, l_lnno);
            }

        }

    }

    nsysms_min = i32min(coff1->filehdr.f_nsyms, coff2->filehdr.f_nsyms);
    for (i32 = 0; i32 < nsysms_min; i32++) {
        c = strcmp(coff1->syments[i32].name, coff2->syments[i32].name);
        if (c) {
            printf("%s %s differ in symbol name %d\n",
                coff1->name, coff2->name, i32);
            if (!keep_going)
                return c;
        }

        coff_cmp1(syments, i32, syment.n_value);
        coff_cmp1(syments, i32, syment.n_scnum);
        coff_cmp1(syments, i32, syment.n_type);
        coff_cmp1(syments, i32, syment.n_sclass);
        coff_cmp1(syments, i32, syment.n_numaux);
    }

    return 0;
}
