#ifndef _BW_HT_H
#define _BW_HT_H 1

#define _GNU_SOURCE
#include <stddef.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdint.h>
#include <inttypes.h>
#include <stdbool.h>
#include <math.h>
#include <errno.h>

#include "bw.h"
#include "cmp.h"
#include "st.h"

struct bw_ht;
struct bw_hte;

typedef void   (*bw_hte_free_fn_t)(struct bw_hte *);
typedef size_t (*bw_hte_hash_fn_t)(const struct bw_hte *);

typedef struct bw_ht_ops {
    bw_iseq_fn_t     iseq_fn;
    bw_hte_hash_fn_t hash_fn;
    bw_hte_free_fn_t free_fn;
} bw_ht_ops_t;

typedef struct bw_hte {
    bw_st_t   s;
    size_t    h;
} bw_hte_t;

typedef struct bw_ht {
    size_t      s, n;
    bw_ht_ops_t *o;
    bw_st_t     **t;
} bw_ht_t;

static bw_hte_t *
_bw_s2e(bw_st_t *s)
{
    return container_of(s, bw_hte_t, s);
}

static bw_st_t *
_bw_e2s(bw_hte_t *e)
{
    return &e->s;
}

static void
_bw_hte_free_fn(bw_st_t *s, void *ctx)
{
    bw_hte_t *e = _bw_s2e(s);
    bw_ht_t *ht = ctx;

    e->h = 0;
    if (ht->o->free_fn)
        ht->o->free_fn(e);
}

static bw_hte_t *
_bw_st_get(bw_st_t *st, bw_hte_t *key, bw_ht_ops_t *o)
{
    while (st) {
        bw_hte_t *e = _bw_s2e(st);
        if (key->h == e->h)
            if (o->iseq_fn(e, key))
                return e;
        st = bw_st_next(st);
    }

    return NULL;
}

static void
bw_hte_init(bw_hte_t *hte)
{
    if (!hte)
        return;

    bw_st_init(_bw_e2s(hte));
    hte->h = 0;
}

#define BW_HTE_INIT() { .s = BW_ST_INIT(), .h = 0 }

static int
bw_ht_init_extern(bw_ht_t *ht, size_t size, bw_ht_ops_t *ops, void *buf)
{
    if (!ht || !size || !ops || !buf)
        return EINVAL;
    if (!ops->hash_fn || !ops->iseq_fn)
        return EINVAL;

    memset(ht, 0, sizeof(*ht));
    ht->t = buf;
    memset(ht->t, 0, size * sizeof(*ht->t));
    ht->o = ops;
    ht->s = size;

    return ESUCCESS;
}

static void
bw_ht_reset(bw_ht_t *ht)
{
    bw_ht_init_extern(ht, ht->s, ht->o, ht->t);
}

static int
bw_ht_init(bw_ht_t *ht, size_t size, bw_ht_ops_t *ops)
{
    bw_st_t **t = NULL;
    int ret;

    if (!ht)
        return EINVAL;
    if (!size)
        return EINVAL;

    t = malloc(size * sizeof(*ht->t));
    if (!t)
        /* malloc sets errno */
        return errno;

    ret = bw_ht_init_extern(ht, size, ops, t);

    if (ret)
        free(t);

    return ret;
}

static bw_ht_t *
bw_ht_new(size_t size, bw_ht_ops_t *ops)
{
    bw_ht_t *new = NULL;
    int ret;

    if (!size) {
        errno = EINVAL;
        return NULL;
    }

    new = malloc(sizeof(bw_ht_t) + (size * sizeof(*new->t)));
    if (!new)
        /* malloc sets errno */
        goto out;

    ret = bw_ht_init_extern(new, size, ops, (char *)new + sizeof(*new));
    if (ESUCCESS != ret) {
        errno = ret;
        goto out_free;
    }

    return new;

out_free:
    free(new);
out:
    return NULL;
}

static void *
bw_ht_clear_extern(bw_ht_t *ht)
{
    size_t i;
    void *buf = NULL;

    if (!ht) {
        errno = EINVAL;
        return NULL;
    }

    if (ht->t) {
        for (i = 0; i < ht->s; i++) {
            errno = bw_st_free(&ht->t[i], _bw_hte_free_fn, ht);
            if (errno != ESUCCESS)
                break;
        }
        buf = ht->t;
    }

    memset(ht, 0, sizeof(*ht));

    return buf;
}

static int
bw_ht_clear(bw_ht_t *ht)
{
    if (!ht)
        return EINVAL;

    if (ht->t)
        free(bw_ht_clear_extern(ht));
    else
        bw_ht_clear_extern(ht);

    return errno;
}

static int
bw_ht_free(bw_ht_t *ht)
{
    if (!ht)
        return EINVAL;

    /* ignore returned pointer, it belongs to `ht` */
    bw_ht_clear_extern(ht);

    free(ht);

    return errno;
}

static size_t
bw_ht_size(bw_ht_t *ht)
{
    if (!ht) {
        errno = EINVAL;
        return 0;
    }

    return ht->n;
}

static bool
bw_ht_empty(bw_ht_t *ht)
{
    if (!ht) {
        errno = EINVAL;
        return NULL;
    }

    return ht->n ? false : true;
}

static int
bw_ht_add(bw_ht_t *ht, bw_hte_t *new)
{
    size_t idx;
    bw_hte_t *old = NULL;

    if (!ht || !new)
        return EINVAL;

    new->h = ht->o->hash_fn(new);
    idx = new->h % ht->s;

    old = _bw_st_get(ht->t[idx], new, ht->o);

    if (!old) {
        bw_st_push(&ht->t[idx], _bw_e2s(new));
        ht->n++;
    } else
        new->h = 0;

    return !old ? ESUCCESS : EAGAIN;
}

static bw_hte_t *
bw_ht_addr(bw_ht_t *ht, bw_hte_t *new)
{
    size_t idx;
    bw_hte_t *old = NULL;

    if (!ht || !new) {
        errno = EINVAL;
        return NULL;
    }

    new->h = ht->o->hash_fn(new);
    idx = new->h % ht->s;

    old = _bw_st_get(ht->t[idx], new, ht->o);

    if (old) {
        bw_st_del(&ht->t[idx], _bw_e2s(old));
        old->h = 0;
    } else
        ht->n++;

    bw_st_push(&ht->t[idx], _bw_e2s(new));

    return old;
}

static bw_hte_t *
bw_ht_get(bw_ht_t *ht, bw_hte_t *key)
{
    size_t idx;

    if (!ht || !key) {
        errno = EINVAL;
        return NULL;
    }

    key->h = ht->o->hash_fn(key);
    idx = key->h % ht->s;

    return _bw_st_get(ht->t[idx], key, ht->o);
}

static int
bw_ht_del(bw_ht_t *ht, bw_hte_t *old)
{
    size_t idx;

    if (!ht || !old)
        return EINVAL;

    idx = old->h % ht->s;

    bw_st_del(&ht->t[idx], _bw_e2s(old));
    old->h = 0;

    ht->n--;
    return ESUCCESS;
}

/* simple iterator interface */
#define bw_ht_forall(__e, __ht, __m, ...)                           \
({                                                                  \
    bw_ht_t *__ht_fa_ht = (__ht);                                   \
    size_t __ht_fa_i;                                               \
    for (__ht_fa_i = 0; __ht_fa_i < (__ht_fa_ht)->s; __ht_fa_i++) { \
        bw_hte_t *__ht_fa_hte;                                      \
        bw_st_forall (__ht_fa_hte, (__ht_fa_ht)->t[__ht_fa_i], s, { \
            __e = container_of(__ht_fa_hte, typeof(*(__e)), __m);   \
            (__VA_ARGS__);                                          \
        })                                                          \
    }                                                               \
});

#endif /* _BW_HT_H */
