#ifndef _SUBPROC_H
#define _SUBPROC_H

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include <inttypes.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <errno.h>
#include <ctype.h>

typedef int (*subproc_fn_t)(void *);

static pid_t
subproc_open(subproc_fn_t subproc_fn, void *arg, int *rdfd, int *wrfd)
{
    int     pfd1[2];
    int     pfd2[2];
    pid_t   pid;

    if (subproc_fn == NULL || rdfd == NULL || wrfd == NULL) {
        errno = EINVAL;
        goto out;
    }

    if (pipe(pfd1) < 0)
        goto out;

    if (pipe(pfd2) < 0)
        goto out_close1;

    pid = fork();
    switch (pid) {

        case -1:
            goto out_close2;

        /* Child process */
        case 0:
            close(pfd1[1]);
            close(pfd2[0]);

            if (pfd1[0] != fileno(stdin))
                dup2(pfd1[0], fileno(stdin));
            close(pfd1[0]);

            if (pfd2[1] != fileno(stdout))
                dup2(pfd2[1], fileno(stdout));
            close(pfd2[1]);

            setlinebuf(stdin);
            setlinebuf(stdout);
            _exit(subproc_fn(arg));

        /* Parent process */
        default:
            close(pfd1[0]);
            close(pfd2[1]);

            *rdfd = pfd2[0];
            *wrfd = pfd1[1];

            return pid;
    }

out_close2:
    close(pfd2[0]);
    close(pfd2[1]);
out_close1:
    close(pfd1[0]);
    close(pfd1[1]);
out:
    return (pid_t)-1;
}

static pid_t
subproc_fopen(subproc_fn_t subproc_fn, void *arg, FILE **rd, FILE **wr)
{
    pid_t pid;
    int rdfd, wrfd;

    if (rd == NULL || wr == NULL)
        return (pid_t)-1;

    pid = subproc_open(subproc_fn, arg, &rdfd, &wrfd);
    if (pid != (pid_t)-1) {
        *rd = fdopen(rdfd, "r");
        setlinebuf(*rd);
        *wr = fdopen(wrfd, "w");
        setlinebuf(*wr);
    }

    return pid;
}

static int
subproc_exec(void *arg)
{
    char **argv = arg;
    execv(argv[0], argv);
    return 127;
}

static int
subproc_execvp(void *arg)
{
    char **argv = arg;
    execvp(argv[0], argv);
    return 127;
}

static pid_t
popenrw(const char *job, FILE **rd, FILE **wr)
{
    const char *shell[] = {
        "/bin/bash", "-c", job, NULL
    };

    return subproc_fopen(subproc_exec, shell, rd, wr);
}

static int
subproc_close(pid_t pid, int rdfd, int wrfd)
{
    int ret;

    if (wrfd >= 0)
        close(wrfd);
    if (rdfd >= 0)
        close(rdfd);

    while (wait4(pid, &ret, 0, NULL) < 0)
        if (errno != EINTR)
            return -1;

    return ret;
}

static int
subproc_fclose(pid_t pid, FILE *rd, FILE *wr)
{
    if (rd)
        fclose(rd);
    if (wr)
        fclose(wr);

    return subproc_close(pid, -1, -1);
}

static int
pcloserw(pid_t pid, FILE *rd, FILE *wr)
{
    return subproc_fclose(pid, rd, wr);
}

#endif /* _SUBPROC_H */
