#define _GNU_SOURCE
#include "config.h"
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdint.h>
#include <stdbool.h>
#include <math.h>
#include <assert.h>
#include <bw/ht.h>

#include "processor.h"
#include "glovars.h"
#include "simul.h"

#include "mem.h"

enum {
	STAT_READ  = 1,
	STAT_WRITE = 2
};

static size_t memStatsHash(const bw_hte_t *hte)
{
	struct vpregs *vp = container_of(hte, struct vpregs, stats.hte);

	return (size_t)vp->stats.addr;
}

static bool memStatsIseq(const void *hte1, const void *hte2)
{
	struct vpregs *vp1 = container_of(hte1, struct vpregs, stats.hte);
	struct vpregs *vp2 = container_of(hte2, struct vpregs, stats.hte);

	return vp1->stats.addr == vp2->stats.addr;
}

char *MemStatsFilename = NULL;
static bw_ht_t memStats;
static FILE *memStatsFile;

static bw_ht_ops_t memStatsOps = {
	.iseq_fn = memStatsIseq,
	.hash_fn = memStatsHash
};

void InitMemStats(void)
{
	bool statsFilenameGenerated = false;

	if (!MemStatsFilename)
		return;

	if (!MemStatsFilename[0]
	    || !strcmp(MemStatsFilename, ":generate:")) {
		MemStatsFilename = GenerateFilename("stats");
		statsFilenameGenerated = true;
	}

	bw_ht_init(&memStats, (PPnum*VPnum) >> 4 ?: 4, &memStatsOps);
	memStatsFile = fopen(MemStatsFilename, "w");

	if (statsFilenameGenerated) {
		free(MemStatsFilename);
		MemStatsFilename = ":generate:";
	}
}

void FiniMemStats(void)
{
	if (memStatsFile)
		fclose(memStatsFile);
	memStatsFile = NULL;
}

void ResetStatMemRequest(struct vpregs *vp)
{
	bw_hte_init(&vp->stats.hte);
	vp->stats.count = 0;
	vp->stats.type  = 0;
}

void DoStatMemRequest(struct vpregs *vp)
{
	struct memreq *mr = &vp->memreq;
	bw_hte_t *hte = NULL;

	if (!memStatsFile || NEGATIV(mr->vaddr))
		return;

	/* only shared access are going into the stats */
	vp->stats.addr = mr->vaddr;

	hte = bw_ht_get(&memStats, &vp->stats.hte);
	if (!hte) {
		/* vp is our new representative for this memory location */
		bw_ht_add(&memStats, &vp->stats.hte);
	} else {
		/* use the available representative for this memory location */
		vp = container_of(hte, struct vpregs, stats.hte);
	}

	/* add the current vp to the vaddr, count is initialized with zero */
	vp->stats.count++;
}

void MarkReadStatMemRequest(struct vpregs *vp)
{
	struct memreq *mr = &vp->memreq;
	bw_hte_t *hte = NULL;

	if (!memStatsFile || NEGATIV(mr->vaddr))
		return;

	hte = bw_ht_get(&memStats, &vp->stats.hte);
	if (!hte) {
		return;
	} else {
		vp = container_of(hte, struct vpregs, stats.hte);
	}

	vp->stats.type |= STAT_READ;
}

void MarkWriteStatMemRequest(struct vpregs *vp)
{
	struct memreq *mr = &vp->memreq;
	bw_hte_t *hte = NULL;

	if (!memStatsFile || NEGATIV(mr->vaddr))
		return;

	hte = bw_ht_get(&memStats, &vp->stats.hte);
	if (!hte) {
		return;
	} else {
		vp = container_of(hte, struct vpregs, stats.hte);
	}

	vp->stats.type |= STAT_WRITE;
}

void OutputMemStats(void)
{
	struct vpregs *vp;

	if (!memStatsFile || !bw_ht_size(&memStats))
		return;

	fprintf(memStatsFile, "%08u:%zu",
	        PhysicalProcessor->CT,
	        bw_ht_size(&memStats));

	bw_ht_forall (vp, &memStats, stats.hte, {
		fprintf(memStatsFile, ":%c%c:%"PRIuMAX,
		        vp->stats.type & STAT_READ  ? 'r' : '-',
		        vp->stats.type & STAT_WRITE ? 'w' : '-',
		        vp->stats.count);
	})

	fprintf(memStatsFile, "\n");

	bw_ht_reset(&memStats);
}
